/*
 * Copyright (c) 1993 by Sun Microsystems Inc.
 */

#pragma	ident	"@(#)XXX.c	1.3	94/04/14 SMI"

/*
 * XXX pseudo-device to support I/O to kernel heap.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/buf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/kmem.h>
#include <sys/poll.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#define	DEFAULT_MAXPHYS	(63 * 1024)			/* '126b' */
#define	DEFAULT_SIZE	((u_long)(1024*DEV_BSIZE))	/* 1/2 a meagrebyte */

/*
 * Patch these in the /etc/system file
 */
static int ramdisk_maxphys = DEFAULT_MAXPHYS;
static int ramdisk_size = DEFAULT_SIZE;

/*
 * The entire state of each ramdisk device.
 */
typedef struct {
	char		*ram;		/* the memory we use */
	int		ramsize;	/* how much is there */
	int		maxphys;	/* how big per transfer */
	dev_info_t	*dip;		/* my devinfo handle */
} XXX_devstate_t;

/*
 * An opaque handle where our set of ramdisk devices lives
 */
static void *XXX_state;

static int XXX_open(dev_t *devp, int flag, int otyp, cred_t *cred);
static int XXX_strategy(struct buf *bp);
static int XXX_read(dev_t dev, struct uio *uiop, cred_t *credp);
static int XXX_write(dev_t dev, struct uio *uiop, cred_t *credp);
static int XXX_print(dev_t dev, char *str);

static struct cb_ops XXX_cb_ops = {
	XXX_open,
	nulldev,	/* close */
	XXX_strategy,
	XXX_print,
	nodev,		/* dump */
	XXX_read,
	XXX_write,
	nodev,		/* ioctl */
	nodev,		/* devmap */
	nodev,		/* mmap */
	nodev,		/* segmap */
	nochpoll,	/* poll */
	ddi_prop_op,
	NULL,
	D_NEW | D_MP
};

static int XXX_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
    void **result);
static int XXX_identify(dev_info_t *dip);
static int XXX_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int XXX_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

static struct dev_ops XXX_ops = {
	DEVO_REV,
	0,
	XXX_getinfo,
	XXX_identify,
	nulldev,	/* probe */
	XXX_attach,
	XXX_detach,
	nodev,		/* reset */
	&XXX_cb_ops,
	(struct bus_ops *)0
};


extern struct mod_ops mod_driverops;

static struct modldrv modldrv = {
	&mod_driverops,
	"ramdisk driver v1.6",
	&XXX_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	0
};

int
_init(void)
{
	int e;

	if ((e = ddi_soft_state_init(&XXX_state,
	    sizeof (XXX_devstate_t), 1)) != 0) {
		return (e);
	}

	if ((e = mod_install(&modlinkage)) != 0)  {
		ddi_soft_state_fini(&XXX_state);
	}

	return (e);
}

int
_fini(void)
{
	int e;

	if ((e = mod_remove(&modlinkage)) != 0)  {
		return (e);
	}
	ddi_soft_state_fini(&XXX_state);
	return (e);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

static int
XXX_identify(dev_info_t *dip)
{
	if (strcmp(ddi_get_name(dip), "ramdisk") == 0) {
		return (DDI_IDENTIFIED);
	} else
		return (DDI_NOT_IDENTIFIED);
}

static int
XXX_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int nblocks;
	int instance;
	char rawname[15], blkname[12];
	XXX_devstate_t *rsp;

	switch (cmd) {

	case DDI_ATTACH:

		instance = ddi_get_instance(dip);

		if (ddi_soft_state_zalloc(XXX_state, instance) != DDI_SUCCESS) {
			cmn_err(CE_CONT, "%s%d: can't allocate state\n",
			    ddi_get_name(dip), instance);
			return (DDI_FAILURE);
		} else
			rsp = ddi_get_soft_state(XXX_state, instance);

		rsp->maxphys = ramdisk_maxphys;
		rsp->ramsize = ramdisk_size;

		rsp->ram = kmem_alloc(rsp->ramsize, KM_NOSLEEP);
		if (rsp->ram == (void *)0) {
			cmn_err(CE_CONT,
			    "%s%d: can't allocate %d bytes for disk\n",
			    ddi_get_name(dip), instance, rsp->ramsize);
			goto attach_failed;
		}

		/*
		 * The 'nblocks' property is mandatory for block devices
		 */
		nblocks = rsp->ramsize / DEV_BSIZE;
		if (ddi_prop_create(DDI_DEV_T_NONE, dip, DDI_PROP_CANSLEEP,
		    "nblocks", (caddr_t)&nblocks, sizeof (int))
		    != DDI_PROP_SUCCESS) {
			cmn_err(CE_CONT, "%s%d: can't create nblocks prop\n",
			    ddi_get_name(dip), instance);
			goto attach_failed;
		}

		(void) sprintf(rawname, "%d,raw", instance);
		(void) sprintf(blkname, "%d", instance);
		if ((ddi_create_minor_node(dip,
		    rawname, S_IFCHR, instance, NULL, 0) == DDI_FAILURE) ||
		    (ddi_create_minor_node(dip,
		    blkname, S_IFBLK, instance, NULL, 0) == DDI_FAILURE)) {
			ddi_remove_minor_node(dip, NULL);
			goto attach_failed;
		}

		rsp->dip = dip;
		ddi_report_dev(dip);
		cmn_err(CE_CONT, "%s%d: misusing %d bytes of memory\n",
			ddi_get_name(dip), ddi_get_instance(dip),
			rsp->ramsize);
		return (DDI_SUCCESS);

	default:
		return (DDI_FAILURE);
	}

attach_failed:
	/*
	 * Use our own detach routine to toss
	 * away any stuff we allocated above.
	 */
	(void) XXX_detach(dip, DDI_DETACH);
	return (DDI_FAILURE);
}

static int
XXX_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int instance;
	register XXX_devstate_t *rsp;

	switch (cmd) {

	case DDI_DETACH:
		/*
		 * Undo what we did in XXX_attach, freeing resources
		 * and removing things we installed.  The system
		 * framework guarantees we are not active with this devinfo
		 * node in any other entry points at this time.
		 */
		ddi_prop_remove_all(dip);
		instance = ddi_get_instance(dip);
		rsp = ddi_get_soft_state(XXX_state, instance);
		if (rsp->ram && rsp->ramsize)
			kmem_free(rsp->ram, rsp->ramsize);
		ddi_remove_minor_node(dip, NULL);
		ddi_soft_state_free(XXX_state, instance);
		return (DDI_SUCCESS);

	default:
		return (DDI_FAILURE);
	}
}

/*ARGSUSED*/
static int
XXX_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	XXX_devstate_t *rsp;
	int error = DDI_FAILURE;

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		if ((rsp = ddi_get_soft_state(XXX_state,
		    getminor((dev_t) arg))) != NULL) {
			*result = rsp->dip;
			error = DDI_SUCCESS;
		} else
			*result = NULL;
		break;

	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)getminor((dev_t) arg);
		error = DDI_SUCCESS;
		break;

	default:
		break;
	}

	return (error);
}


/*ARGSUSED*/
static int
XXX_open(dev_t *devp, int flag, int otyp, cred_t *cred)
{
	if (otyp != OTYP_BLK && otyp != OTYP_CHR)
		return (EINVAL);

	if (ddi_get_soft_state(XXX_state, getminor(*devp)) == NULL)
		return (ENXIO);

	return (0);
}

static void
XXX_minphys(struct buf *bp)
{
	XXX_devstate_t *rsp;

	rsp = ddi_get_soft_state(XXX_state, getminor(bp->b_edev));
	if (bp->b_bcount > rsp->maxphys)
		bp->b_bcount = rsp->maxphys;
}

/*ARGSUSED*/
static int
XXX_read(dev_t dev, struct uio *uiop, cred_t *credp)
{
	int instance = getminor(dev);
	XXX_devstate_t *rsp = ddi_get_soft_state(XXX_state, instance);

	if (uiop->uio_offset >= rsp->ramsize)
		return (EINVAL);

	return (physio(XXX_strategy, (struct buf *)0, dev, B_READ,
	    XXX_minphys, uiop));
}

/*ARGSUSED*/
static int
XXX_write(dev_t dev, register struct uio *uiop, cred_t *credp)
{
	int instance = getminor(dev);
	XXX_devstate_t *rsp = ddi_get_soft_state(XXX_state, instance);

	if (uiop->uio_offset >= rsp->ramsize)
		return (EINVAL);

	return (physio(XXX_strategy, (struct buf *)0, dev, B_WRITE,
	    XXX_minphys, uiop));
}

static int
XXX_strategy(struct buf *bp)
{
	XXX_devstate_t *rsp;
	u_long offset = bp->b_blkno * DEV_BSIZE;

	rsp = ddi_get_soft_state(XXX_state, getminor(bp->b_edev));
	if (rsp == NULL) {
		bp->b_error = ENXIO;
		bp->b_flags |= B_ERROR;
	} else if (offset >= rsp->ramsize) {
		bp->b_error = EINVAL;
		bp->b_flags |= B_ERROR;
	} else {
		register caddr_t	buf_addr, raddr;
		register unsigned	nbytes;

		raddr = rsp->ram + offset;
		nbytes = min(bp->b_bcount, rsp->ramsize - offset);

		if (bp->b_flags & B_PAGEIO)
			bp_mapin(bp);

		buf_addr = bp->b_un.b_addr;

		if (bp->b_flags & B_READ)
			(void) bcopy(raddr, buf_addr, nbytes);
		else
			(void) bcopy(buf_addr, raddr, nbytes);

		bp->b_resid = bp->b_bcount - nbytes;

		if (bp->b_flags & B_PAGEIO)
			bp_mapout(bp);
	}

	biodone(bp);
	return (0);
}

static int
XXX_print(dev_t dev, char *str)
{
	int instance = getminor(dev);
	XXX_devstate_t *rsp = ddi_get_soft_state(XXX_state, instance);

	cmn_err(CE_WARN, "%s%d: %s\n", ddi_get_name(rsp->dip), instance, str);
	return (0);
}
